"""
A wrapper of the subprocess module

This module relies on the following system utilities being installed:
* bash
* kill
* pstree
* tee
"""

import io
import os
import re
import sys
import shlex
import psutil
import subprocess

from subprocess import DEVNULL
from subprocess import PIPE
from subprocess import STDOUT
from subprocess import CompletedProcess
from subprocess import TimeoutExpired
from subprocess import SubprocessError  # noqa: F401
from subprocess import CalledProcessError

from time import time
from typing import Union, Iterable, Callable
from tempfile import SpooledTemporaryFile
from contextlib import contextmanager, nullcontext

FILE = -4
GLOBAL_ECHO = False
GLOBAL_STDOUTS = set()
GLOBAL_STDERRS = set()
GLOBAL_PROMPTS = set()


TeeStream = Union[io.IOBase, Callable[[], io.IOBase]]


class PipeFile(SpooledTemporaryFile):
    """ A file to be used as stdin, stdout and stderr in Popen to avoid dead lock
        when the process output is too long using PIPE.

        If used as stdin, the content should be written before spawning the process.
    """

    def __init__(self, content=None, encoding=None, errors=None, text=None):
        super().__init__(
            mode="w+t" if text else "w+b",
            encoding=encoding,
            errors=errors
        )

        if content:
            self.write(content)
            self.seek(0)

    def read_all(self):
        self.seek(0)
        return self.read()


class Tee(subprocess.Popen):
    """ A subprocess to send real-time input to several file descriptors """

    def __init__(self, input, fds, output=None, encoding=None, errors=None, text=None):
        if output is STDOUT:
            raise ValueError("output cannot be STDOUT")

        if output is None:
            output = DEVNULL

        fds = normalize_outerr_fds(fds)
        if 1 in fds:
            if output != DEVNULL:
                fds.add(2)
            stdout = None
            stderr = output
        elif 2 in fds:
            stdout = output
            stderr = None
        else:
            stdout = output
            stderr = DEVNULL

        fds.discard(1)
        super().__init__(
            stdin=input, stdout=stdout, stderr=stderr, pass_fds=fds - {2},
            encoding=encoding, errors=errors, text=text, **self.get_kwargs(fds)
        )

        self.output = self.stdout if self.stderr is None else self.stderr

    @staticmethod
    def get_cmd(fds):
        return ["tee", "-a"] + [f"/dev/fd/{fd}" for fd in fds]

    if os.access("/dev/fd/2", os.W_OK):
        @classmethod
        def get_kwargs(cls, fds):
            return {"args": cls.get_cmd(fds)}
    else:
        @classmethod
        def get_kwargs(cls, fds):
            if 2 in fds:
                return {
                    "args": " ".join(cls.get_cmd(fds - {2})) + " >(cat >&2)",
                    "shell": True
                }
            return {"args": cls.get_cmd(fds)}

    def communicate(self, *args, **kwargs):
        stdout, stderr = super().communicate(*args, **kwargs)
        return stdout if stderr is None else stderr


class Popen(subprocess.Popen):
    """ A subprocess.Popen that can send its stdout and stderr to several files
        in real-time keeping the ability of capturing its output.
    """

    def __init__(
        self, args, *, cwd=None, env=None, encoding=None, errors=None, text=None,
        stdout_tees: Iterable[TeeStream] = [], add_global_stdout_tees=True,
        stderr_tees: Iterable[TeeStream] = [], add_global_stderr_tees=True,
        prompt_tees: Iterable[TeeStream] = [], add_global_prompt_tees=True,
        echo=None, alias=None, comment=None, **kwargs
    ):
        stdout = kwargs.get("stdout")
        stderr = kwargs.get("stderr")

        stdout_tees, stderr_tees, prompt_tees = self._get_tee_sets(
            stdout_tees, add_global_stdout_tees,
            stderr_tees, add_global_stderr_tees,
            prompt_tees, add_global_prompt_tees,
            echo, stdout, stderr
        )

        if stderr is STDOUT:
            err2out = True
            stderr_tees = set()
        else:
            err2out = False

        stdout_fds = {tee.fileno() for tee in stdout_tees}
        stderr_fds = {tee.fileno() for tee in stderr_tees}
        prompt_fds = {tee.fileno() for tee in prompt_tees}

        if stdout_fds:
            kwargs["stdout"] = PIPE

        if stderr_fds:
            kwargs["stderr"] = PIPE

        if prompt_fds:
            stream_prompts(prompt_fds, alias or args, cwd, env, err2out, comment)

        super().__init__(args, cwd=cwd, env=env,
                         encoding=encoding, errors=errors, text=text, **kwargs)

        self.original_stdout = self.stdout
        self.original_stderr = self.stderr

        stdout_process = stderr_process = self
        if stdout_fds:
            stdout_process = Tee(self.stdout, stdout_fds, stdout, encoding, errors, text)
            self.stdout = stdout_process.output
        if stderr_fds:
            stderr_process = Tee(self.stderr, stderr_fds, stderr, encoding, errors, text)
            self.stderr = stderr_process.output

        self.stdout_process = stdout_process
        self.stderr_process = stderr_process

    @staticmethod
    def _get_tee_files(tees: Iterable[TeeStream]):
        return set(callable(tee) and tee() or tee for tee in tees)

    @classmethod
    def _get_tee_sets(
        cls,
        stdout_tees, add_global_stdout_tees,
        stderr_tees, add_global_stderr_tees,
        prompt_tees, add_global_prompt_tees,
        echo, stdout, stderr
    ):
        stdout_tees = set(stdout_tees)
        stderr_tees = set(stderr_tees)
        prompt_tees = set(prompt_tees)

        if add_global_prompt_tees:
            prompt_tees.update(GLOBAL_PROMPTS)
        if add_global_stdout_tees:
            stdout_tees.update(GLOBAL_STDOUTS)
        if add_global_stderr_tees:
            stderr_tees.update(GLOBAL_STDERRS)

        if echo is None:
            echo = GLOBAL_ECHO

        if echo:
            prompt_tees.add(sys.stdout)
            stdout_tees.add(sys.stdout)
            stderr_tees.add(sys.stderr)

        if stdout is None:
            if stdout_tees == {sys.stdout}:
                stdout_tees.clear()  # tee process not needed for stdout
            if stdout_tees:
                stdout_tees.add(sys.stdout)

        if stderr is None:
            if stderr_tees == {sys.stderr}:
                stderr_tees.clear()  # tee process not needed for stderr
            if stderr_tees:
                stderr_tees.add(sys.stderr)

        return (
            cls._get_tee_files(stdout_tees),
            cls._get_tee_files(stderr_tees),
            cls._get_tee_files(prompt_tees)
        )

    def get_pids(self):
        """ Return the pid for all of the processes in the tree """
        output = run(
            ["pstree", "-p", str(self.pid)],
            stdin=DEVNULL,
            stdout=PIPE,
            stderr=DEVNULL,
            text=True,
            check=True
        )

        return re.findall("\\((\\d+)\\)", output.stdout)[::-1]

    def terminate_tree(self):
        run(
            ["kill"] + self.get_pids(),
            stdin=DEVNULL, stdout=DEVNULL, stderr=DEVNULL,
        )

    def kill_tree(self):
        run(
            ["kill", "-9"] + self.get_pids(),
            stdin=DEVNULL, stdout=DEVNULL, stderr=DEVNULL,
        )

    def end_tree(self, sigterm_timeout):
        """ Try to gracefully terminate the process tree,
            kill it after 'sigterm_timeout' seconds
        """
        if sigterm_timeout:
            self.terminate_tree()
            try:
                self.communicate(timeout=sigterm_timeout)
            except:  # noqa: E722
                self.kill_tree()
        else:
            self.kill_tree()

    def poll(self):
        processes = {self.stderr_process, self.stdout_process} - {self}
        if all(process.poll() is not None for process in processes):
            return super().poll()

    def wait(self, timeout=None):
        processes = {self.stderr_process, self.stdout_process} - {self}
        for process in processes:
            process.wait(timeout)
            timeout = None
        return super().wait(timeout)

    @contextmanager
    def _streams_restored(self):
        self.stdout = self.original_stdout
        self.stderr = self.original_stderr
        try:
            yield
        finally:
            self.stdout = self.stdout_process.stdout
            self.stderr = self.stderr_process.stderr

    @contextmanager
    def _stdin_none(self):
        original_stdin = self.stdin
        self.stdin = None
        try:
            yield
        finally:
            self.stdin = original_stdin

    def _communicate_tees(self, timeout):
        stdout = stderr = None
        if self.stdout_process is not self:
            stdout = self.stdout_process.communicate(timeout=timeout)
            timeout = None
        if self.stderr_process is not self:
            stderr = self.stderr_process.communicate(timeout=timeout)
            timeout = None
        return stdout, stderr, timeout

    def _communicate_all(self, timeout):
        stdout, stderr, timeout = self._communicate_tees(timeout)
        main_stdout, main_stderr = super().communicate(timeout=timeout)
        return (
            main_stdout if self.stdout_process is self else stdout,
            main_stderr if self.stderr_process is self else stderr
        )

    @property
    def _any_communication_started(self):
        return (
            self.stdout_process._communication_started or
            self.stderr_process._communication_started or
            self._communication_started
        )

    def communicate(self, input=None, timeout=None):
        if self._any_communication_started and input:
            raise ValueError("Cannot send input after starting communication")

        with self._streams_restored():
            if self.stdin and input:
                self._stdin_write(input)
                with self._stdin_none():
                    return self._communicate_all(timeout)
            return self._communicate_all(timeout)


def normalize_outerr_fds(fds: Iterable[int]):
    """ Return fds as a set but using 1 and 2 for stdout and stderr file
        descriptors in case we are being redirected
    """
    out_fd = sys.stdout.fileno()  # This might not always be 1
    err_fd = sys.stderr.fileno()  # This might not always be 2
    fds = set(fds)
    if out_fd in fds:
        fds.remove(out_fd)
        fds.add(1)
    if err_fd in fds:
        fds.remove(err_fd)
        fds.add(2)
    return fds


def quote(cmd: Iterable[str]):
    """ Convert the command tokens into a single string that could be pasted into
        the shell to execute the original command
    """
    return " ".join(map(shlex.quote, cmd))


def shellify(cmd: Union[str, Iterable[str]], err2out=False, comment=None):
    """ Quote command if needed and optionally add extra strings to express
        stderr being redirected to stdout and a comment
    """
    if not isinstance(cmd, str):
        cmd = quote(cmd)
    if err2out:
        cmd += " 2>&1"
    if comment:
        cmd += f" # {comment}"
    return cmd


# If bash is installed and supports prompt expansion
if subprocess.run(
    ["bash", "-c", "echo ${0@P}"],
    stdin=DEVNULL,
    stdout=DEVNULL,
    stderr=DEVNULL
).returncode == 0:
    HOME = os.path.expanduser("~")
    PS1, PS2 = subprocess.run(
        ["bash", "-ic", "echo \"$PS1\"; echo \"$PS2\""],
        text=True,
        stdin=DEVNULL,
        stdout=PIPE,
        stderr=DEVNULL
    ).stdout.splitlines()[-2:]

    def stream_prompts(fds: Iterable[str], cmd, cwd=None, env=None, err2out=False, comment=None):
        """ Write shell prompt and command into file descriptors fds """
        fds = normalize_outerr_fds(fds)
        custom_env = {"CPS1": PS1, "CPS2": PS2}
        custom_env.update(env or {})
        custom_env.setdefault("HOME", HOME)
        script = (
            "(\n"
            "    IFS= read -r \"line\"\n"
            "    echo \"${CPS1@P}${line}\"\n"
            "    while IFS= read line; do\n"
            "        echo \"${CPS2@P}${line}\"\n"
            "    done\n"
            ") | " + quote(Tee.get_cmd(fds - {1}))
        )
        subprocess.run(
            ["bash", "-c", script],
            text=True,
            input=shellify(cmd, err2out, comment),
            stdout=None if 1 in fds else DEVNULL,
            stderr=None if 2 in fds else DEVNULL,
            pass_fds=fds - {1, 2},
            cwd=cwd,
            env=custom_env,
            check=True
        )
else:  # Use hard-coded PS1 and PS2 strings
    def stream_prompts(fds: Iterable[str], cmd, cwd=None, env=None, err2out=False, comment=None):
        """ Write shell prompt and command into file descriptors fds """
        cmd = shellify(cmd, err2out, comment) + "\n"
        input = "$ " + "> ".join(cmd.splitlines(keepends=True))
        with Tee(PIPE, fds, DEVNULL, text=True) as tee:
            tee.communicate(input=input)


def set_global_echo(value):
    global GLOBAL_ECHO
    GLOBAL_ECHO = bool(value)


def set_global_stdout_files(*files: TeeStream):
    global GLOBAL_STDOUTS
    GLOBAL_STDOUTS = set(*files)


def set_global_stderr_files(*files: TeeStream):
    global GLOBAL_STDERRS
    GLOBAL_STDERRS = set(*files)


def set_global_prompt_files(*files: TeeStream):
    global GLOBAL_PROMPTS
    GLOBAL_PROMPTS = set(*files)


def _output_context(kwargs, key, encoding, errors, text):
    """ Create a PipeFile, store it in kwargs[key] and return it if it is FILE.
        Just return a nullcontext otherwise.
    """
    if kwargs.get(key) is FILE:
        kwargs[key] = PipeFile(encoding=encoding, errors=errors, text=text)
        return kwargs[key]
    return nullcontext()


def run(
    args, *, input=None, capture_output=False, timeout=None, check=False,
    encoding=None, errors=None, text=None, sigterm_timeout=10, **kwargs
):
    """ A subprocess.run that instantiates this module's Popen """
    if input is not None:
        if kwargs.get("stdin") is not None:
            raise ValueError("stdin and input arguments may not both be used.")
        kwargs["stdin"] = PIPE

    if capture_output:
        if kwargs.get("stdout") is not None or kwargs.get("stderr") is not None:
            raise ValueError("stdout and stderr arguments may not be used with capture_output.")
        kwargs["stdout"] = FILE
        kwargs["stderr"] = FILE

    comment = f"timeout={timeout}" if timeout else None
    with (
        _output_context(kwargs, "stdout", encoding, errors, text) as stdout_file,
        _output_context(kwargs, "stderr", encoding, errors, text) as stderr_file,
        Popen(args, encoding=encoding, errors=errors, text=text,
              comment=comment, **kwargs) as process
    ):
        start = psutil.Process(process.pid).create_time()
        try:
            stdout, stderr = process.communicate(input, timeout=timeout)
        except TimeoutExpired:
            process.end_tree(sigterm_timeout)
            process.wait()
            raise
        except:  # noqa: E722
            process.kill_tree()
            raise
        finally:
            end = time()
            returncode = process.poll()

        if stdout_file:
            stdout = stdout_file.read_all()
        if stderr_file:
            stderr = stderr_file.read_all()

        if check and returncode:
            raise CalledProcessError(returncode, process.args, output=stdout, stderr=stderr)

        output = CompletedProcess(process.args, returncode, stdout, stderr)
        output.time = end - start
        return output


def call(*args, **kwargs):
    return run(*args, **kwargs).returncode


def check_call(*args, **kwargs):
    kwargs["check"] = True
    return call(*args, **kwargs)


def check_output(*args, **kwargs):
    kwargs["check"] = True
    kwargs.setdefault("stdout", PIPE)
    return run(*args, **kwargs).stdout


def getoutput(*args, **kwargs):
    return getstatusoutput(*args, **kwargs)[1]


def getstatusoutput(*args, **kwargs):
    kwargs["stdout"] = PIPE
    kwargs["stderr"] = STDOUT
    kwargs["shell"] = True
    kwargs["text"] = True
    output = run(*args, **kwargs)
    return (output.returncode, output.stdout)



"""
Execute commands as Python functions
"""

from shlex import split
from random import choice
from itertools import chain

from iripau.command import host_run


class Command:
    """ Run an executable command as a Python callable """

    def __init__(self, parent, command):
        self._parent = parent
        self._command = parent._mk_command(command)
        self._mk_command = parent._mk_command

    def __getattr__(self, command):
        child = Command(self, command)
        setattr(self, command, child)
        return child

    def __call__(self, *args, **kwargs):
        return self._parent(self._command, *args, **kwargs)


def make_command(command):
    return command.replace("_", "-")


def make_option(option):
    return "--" + option.replace("_", "-"),


class Executable:
    """ Run an executable as a Python callable """

    def __init__(
        self, executable, make_command=make_command, make_option=make_option,
        alias=None, run_args_prefix="_", run_function=None, **kwargs
    ):
        self._run = run_function or host_run
        self._exe = split(executable) if isinstance(executable, str) else executable
        self._alias = split(alias) if isinstance(alias, str) else alias
        self._kwargs = kwargs

        self._prefix = run_args_prefix
        self._mk_option = make_option
        self._mk_command = make_command

    def __getattr__(self, command):
        child = Command(self, command)
        setattr(self, command, child)
        return child

    def __call__(self, *args, **kwargs):
        optionals = chain.from_iterable(
            self._make_arg(self._mk_option(key), value)
            for key, value in kwargs.items()
            if not key.startswith(self._prefix)
        )

        positionals = list(map(str, args))
        optionals = list(optionals)

        kwargs = {
            key[len(self._prefix):]: value
            for key, value in kwargs.items()
            if key.startswith(self._prefix)
        }

        if self._alias:
            kwargs.setdefault("alias", self._alias + positionals + optionals)

        cmd = self._exe + positionals + optionals
        return self._run(cmd, **self._kwargs, **kwargs)

    @staticmethod
    def _is_iterable(value):
        if isinstance(value, (str, bytes)):
            return False
        return hasattr(value, "__iter__")

    @classmethod
    def _make_arg(cls, options, value=None):
        """ Return a list of tokens. Randomly choose a short or long option.
            If a 'value' is given it's appended appropriately.
            If 'value' is iterable, the option will be repeated for each item.
        """
        if cls._is_iterable(value):
            return chain.from_iterable(
                cls._make_arg(options, item)
                for item in value
            )

        if value in {None, False}:
            return []

        option = choice(options)
        if value is True:
            return [option]

        value = str(value)
        if option.startswith("--"):
            return [option + "=" + value]
        return [option, value]



"""
Tests to validate iripau.functools module
"""

import pytest
import operator
import multiprocessing

from time import sleep

from iripau.functools import wait_for
from iripau.functools import retry
from iripau.functools import globalize


class TestFunctools:

    @pytest.mark.parametrize("timeout", [False, True], ids=["successful", "timed_out"])
    @pytest.mark.parametrize("outcome", [False, True])
    def test_wait_for(self, outcome, timeout):
        data = [0]

        operation = operator.gt if outcome else operator.le

        def condition(arg1, arg2):
            sleep(0.4)
            data[0] = data[0] + 1
            return operation(data[0], 3)

        if timeout:
            with pytest.raises(TimeoutError):
                wait_for(
                    condition, "Arg1", "Arg2",
                    _timeout=1,
                    _outcome=outcome,
                    _poll_time=0.5
                )
        else:
            wait_for(
                condition, "Arg1", "Arg2",
                _timeout=2,
                _outcome=outcome,
                _poll_time=0.5
            )

    def test_wait_for_with_stop_condition(self):
        data = [0]

        def condition(arg1, arg2):
            data[0] = data[0] + 1
            return data[0] > 5

        def stop_condition():
            return data[0] > 3

        with pytest.raises(InterruptedError):
            wait_for(
                condition, "Arg1", "Arg2",
                _poll_time=0.5,
                _stop_condition=stop_condition
            )

        assert data[0] == 4

    @pytest.mark.parametrize("use_yield", [False, True], ids=["no_yield", "yield"])
    @pytest.mark.parametrize("success", [True, False], ids=["successful", "unsuccessful"])
    def test_retry(self, success, use_yield):
        data = [0]
        return_value = "SomeReturnValue"

        def regular_function():
            data[0] = data[0] + 1
            assert data[0] >= 3
            return return_value

        def generator_function():
            data[0] = data[0] + 1
            yield
            assert data[0] >= 3
            return return_value

        function = generator_function if use_yield else regular_function
        function = retry(4 if success else 2, AssertionError)(function)

        if success:
            assert return_value == function()
        else:
            with pytest.raises(AssertionError):
                function()

    def test_retry_failure_before_yield(self):
        data = [0]

        @retry(1, AssertionError)
        def function():
            assert False
            yield
            data[0] = 1

        with pytest.raises(AssertionError):
            function()

        assert data[0] == 0

    @pytest.mark.parametrize("fulfilled", [True, False], ids=["fulfilled", "not_fulfilled"])
    def test_retry_with_condition(self, fulfilled):
        data = [0]
        error_message = "expected message" if fulfilled else "wrong message"

        def condition(exception):
            return "expected" in str(exception)

        @retry(2, AssertionError, condition)
        def function():
            data[0] = data[0] + 1
            assert False, error_message

        with pytest.raises(AssertionError):
            function()

        expected_tries = 2 if fulfilled else 1
        assert data[0] == expected_tries

    def test_globalize(self):

        @globalize
        def local_function(x):
            return x + 1

        with multiprocessing.Pool(processes=1) as pool:
            results = pool.map(local_function, [0])

        assert results == [1]



"""
Tests to validate iripau.executable module
"""

from mock import MagicMock, patch, call

from iripau.executable import Command
from iripau.executable import Executable


class TestExecutable:

    def test_command_call(self):
        parent = MagicMock()
        command = "group"

        group = Command(parent, command)
        parent._mk_command.assert_called_once_with(command)

        output = group("list", color="auto")

        parent.assert_called_once_with(
            parent._mk_command.return_value,
            "list",
            color="auto"
        )
        assert output == parent.return_value

    @patch("iripau.executable.Command")
    def test_command_subcommand(self, mock_command):
        parent = MagicMock()
        command = "group"

        group = Command(parent, command)
        group_install = group.install

        mock_command.assert_called_once_with(group, "install")
        assert group_install == group.install
        mock_command.assert_called_once()

    @patch("iripau.executable.choice")
    @patch("iripau.executable.host_run")
    def test_executable_call(self, mock_run, mock_choice):
        executable = "dnf"
        commands_map = {"check_update": "check-update"}
        options_map = {
            "config": ("-c", "--config"),
            "quiet": ("-q", "--quiet"),
            "skip_broken": ("--skip-broken",)
        }

        mock_choice.side_effect = lambda options: options[0]
        dnf = Executable(executable, commands_map.get, options_map.get)
        output = dnf(
            "group", "install", "Development Tools",
            config="/tmp/dnf.conf", quiet=None, skip_broken=True
        )
        assert output == mock_run.return_value

        expected_choice_calls = [
            call(("-c", "--config")),
            call(("--skip-broken",))
        ]

        mock_choice.assert_has_calls(expected_choice_calls)
        tokens = [
            "dnf",
            "group",
            "install",
            "Development Tools",
            "-c", "/tmp/dnf.conf",
            "--skip-broken"
        ]
        mock_run.assert_called_once_with(tokens)

    @patch("iripau.executable.choice")
    @patch("iripau.executable.host_run")
    def test_executable_append_arguments(self, mock_run, mock_choice):
        executable = "docker"
        options_map = {
            "volume": ("-v", "--volume"),
            "name": ("--name",),
            "privileged": ("--privileged",)
        }

        mock_choice.side_effect = lambda options: options[-1]
        docker = Executable(executable, make_option=options_map.get)
        output = docker.run(
            "ubuntu:latest",
            volume=["/tmp/data:/data", "/home/user/repos:/root/Repos:Z"],
            privileged=True,
            name="ubuntu-dev"
        )
        assert output == mock_run.return_value

        expected_choice_calls = [
            call(("-v", "--volume")),
            call(("--privileged",)),
            call(("--name",))
        ]

        mock_choice.assert_has_calls(expected_choice_calls)
        tokens = [
            "docker",
            "run",
            "ubuntu:latest",
            "--volume=/tmp/data:/data",
            "--volume=/home/user/repos:/root/Repos:Z",
            "--privileged",
            "--name=ubuntu-dev"
        ]
        mock_run.assert_called_once_with(tokens)

    @patch("iripau.executable.host_run")
    def test_executable_host_run_args(self, mock_run):
        executable = "dnf"

        dnf = Executable(executable)
        dnf(_timeout=5)

        mock_run.assert_called_once_with(["dnf"], timeout=5)

    @patch("iripau.executable.host_run")
    def test_executable_alias(self, mock_run):
        executable = "ls -al"

        ls = Executable(executable, alias="ll")
        ls("/tmp")

        mock_run.assert_called_once_with(
            ["ls", "-al", "/tmp"],
            alias=["ll", "/tmp"]
        )

    @patch("iripau.executable.host_run")
    def test_executable_default_underscore_to_dash(self, mock_run):
        executable = "git"

        git = Executable(executable)
        git.format_patch(interdiff="feature/v1", no_attach=True, signature_file="some/path")

        mock_run.assert_called_once_with(
            ["git", "format-patch", "--interdiff=feature/v1",
             "--no-attach", "--signature-file=some/path"]
        )



"""
Function utilities
"""

import sys
import uuid
import operator

from functools import wraps
from inspect import getsource, isgeneratorfunction
from time import time, sleep


def wait_for(
    condition, *args,
    _timeout=None, _outcome=True, _poll_time=10, _stop_condition=None, **kwargs
):
    last = time()
    end = _timeout and last + _timeout
    operation = operator.not_ if _outcome else operator.truth
    while operation(condition(*args, **kwargs)):
        if _stop_condition and _stop_condition():
            message = "No reason to keep waiting since the following condition was met:\n{0}"
            raise InterruptedError(message.format(getsource(_stop_condition)))
        now = time()
        if end and now > end:
            message = "The following condition was not {0} after {1} seconds: \n{2}"
            raise TimeoutError(
                message.format(_outcome, _timeout, getsource(condition))
            )
        sleep_time = max(0, _poll_time - (now - last))
        sleep(sleep_time)
        last = now + sleep_time


def retry(tries, exceptions=Exception, retry_condition=None, backoff_time=0):
    """
        Call the decorated function until it succeeds.

        It will only retry if the expected exceptions are risen and the condition
        is fulfilled. To verify the condition is fulfilled, it will be called
        with the caught exception.

        The decorated function can have two parts divided by 'yield', in that
        case the expected exceptions will only be caught in the second part.
    """
    def decorator(function):

        @wraps(function)
        def helper(*args, **kwargs):
            yield
            return function(*args, **kwargs)

        f = function if isgeneratorfunction(function) else helper

        @wraps(function)
        def wrapper(*args, **kwargs):
            t = tries
            while t > 0:
                gen = f(*args, **kwargs)
                next(gen)
                try:
                    next(gen)
                except StopIteration as e:
                    return e.value
                except exceptions as e:
                    if retry_condition is None or retry_condition(e):
                        if t == 1:
                            raise
                        t = t - 1
                        sleep(backoff_time)
                    else:
                        raise
        return wrapper
    return decorator


def globalize(function):
    """
        Make function globally available in the module it was defined so it can
        be serialized. Useful when calling local functions with multiprocessing.
    """
    @wraps(function)
    def wrapper(*args, **kwargs):
        return function(*args, **kwargs)

    wrapper.__name__ = wrapper.__qualname__ = uuid.uuid4().hex
    setattr(sys.modules[function.__module__], wrapper.__name__, wrapper)
    return wrapper



"""
Tests to validate iripau.subprocess module
"""

import sys
import pytest
import subprocess

from mock import patch
from shlex import quote
from tempfile import SpooledTemporaryFile

from iripau.subprocess import DEVNULL
from iripau.subprocess import PIPE
from iripau.subprocess import STDOUT
from iripau.subprocess import FILE
from iripau.subprocess import Tee
from iripau.subprocess import Popen
from iripau.subprocess import TimeoutExpired
from iripau.subprocess import CalledProcessError
from iripau.subprocess import run
from iripau.subprocess import call
from iripau.subprocess import check_call
from iripau.subprocess import check_output
from iripau.subprocess import getoutput
from iripau.subprocess import getstatusoutput

KWARGS = {
    "stdin": DEVNULL,
    "capture_output": True,
    "text": True
}


def get_prompt_and_command(command, err2out=False, timeout=None):
    prompt = subprocess.run(
        "echo '' | bash -i 2>&1 1>/dev/null | head -1",
        shell=True,
        check=True,
        text=True,
        stdout=subprocess.PIPE
    ).stdout[:-1]

    if not isinstance(command, str):
        command = " ".join(map(quote, command))

    if err2out:
        command += " 2>&1"

    if timeout:
        command += f" # timeout={timeout}"

    return f"{prompt}{command}"


def read_file(file):
    file.seek(0)
    return file.read()


class TestSubprocess:

    def test_tee_stdout(self):
        with pytest.raises(ValueError):
            Tee(PIPE, [], STDOUT)

    @pytest.mark.parametrize("echo", [False, True], ids=["no_echo", "echo"])
    def test_popen(self, echo, capfd):
        command = "echo $BASHPID; echo Sup!; sleep 1; echo Bye. >&2"
        stdout_tees = [
            SpooledTemporaryFile(mode="w+t"),
            SpooledTemporaryFile(mode="w+t"),
            SpooledTemporaryFile(mode="w+t")
        ]
        stderr_tees = [
            SpooledTemporaryFile(mode="w+t"),
            SpooledTemporaryFile(mode="w+t"),
            SpooledTemporaryFile(mode="w+t")
        ]
        prompt_tees = [
            SpooledTemporaryFile(mode="w+t"),
            SpooledTemporaryFile(mode="w+t"),
            SpooledTemporaryFile(mode="w+t")
        ]
        all_tee = SpooledTemporaryFile(mode="w+t")

        process = Popen(
            command,
            stdin=DEVNULL,
            stdout=PIPE,
            stderr=PIPE,
            shell=True,
            executable="bash",
            echo=echo,
            stdout_tees=stdout_tees + [all_tee],
            stderr_tees=stderr_tees + [all_tee],
            prompt_tees=prompt_tees + [all_tee],
            text=True
        )

        with process:
            assert f"{process.pid}\n" == process.stdout.readline()
            stdout, stderr = process.communicate()

        assert "Sup!\n" == stdout
        assert "Bye.\n" == stderr

        expected_stdout = f"{process.pid}\nSup!\n"
        for file in stdout_tees:
            assert expected_stdout == read_file(file)

        expected_stderr = "Bye.\n"
        for file in stderr_tees:
            assert expected_stderr == read_file(file)

        expected_prompt = get_prompt_and_command(command) + "\n"
        for file in prompt_tees:
            assert expected_prompt == read_file(file)

        expected_output = expected_prompt + expected_stdout + expected_stderr
        assert expected_output == read_file(all_tee)

        out, err = capfd.readouterr()
        if echo:
            assert expected_prompt + expected_stdout == out
            assert expected_stderr == err
        else:
            assert "" == out == err

    @pytest.mark.parametrize("echo", [False, True], ids=["no_echo", "echo"])
    def test_run_success(self, echo, capfd):
        command = ["echo", "Hello!"]
        output = run(command, echo=echo, **KWARGS)

        assert output.stdout == "Hello!\n"
        assert output.stderr == ""
        assert output.returncode == 0
        assert output.time > 0

        out, err = capfd.readouterr()
        if echo:
            assert get_prompt_and_command(command) + f"\n{output.stdout}" == out
            assert output.stderr == err
        else:
            assert "" == out == err

    def test_run_non_capturing_echo(self, capfd):
        command = ["echo", "Hello!"]
        output = run(command, stdout=None, stderr=None, echo=True)

        assert output.stdout is None
        assert output.stderr is None
        assert output.returncode == 0
        assert output.time > 0

        out, err = capfd.readouterr()
        assert get_prompt_and_command(command) + "\nHello!\n" == out
        assert "" == err

    @pytest.mark.parametrize("check", [False, True], ids=["no_check", "check"])
    @pytest.mark.parametrize("echo", [False, True], ids=["no_echo", "echo"])
    def test_run_error(self, echo, check, capfd):
        command = "echo Hello! >&2; exit 1"
        if check:
            with pytest.raises(CalledProcessError):
                run(command, shell=True, echo=echo, check=True)
        else:
            output = run(command, shell=True, echo=echo, **KWARGS)

            assert output.stdout == ""
            assert output.stderr == "Hello!\n"
            assert output.returncode == 1
            assert output.time > 0

            out, err = capfd.readouterr()
            if echo:
                assert get_prompt_and_command(command) + f"\n{output.stdout}" == out
                assert output.stderr == err
            else:
                assert "" == out == err

    def test_run_time(self):
        command = ["sleep", "3"]
        output = run(command)
        assert output.time > 3

    @pytest.mark.parametrize("echo", [False, True], ids=["no_echo", "echo"])
    def test_run_timeout_and_terminate(self, echo, capfd):
        command = ["sleep", "3"]
        timeout = 1
        with pytest.raises(TimeoutExpired):
            run(command, timeout=timeout, echo=echo)

        if echo:
            out, err = capfd.readouterr()
            assert get_prompt_and_command(command, False, timeout) + "\n" == out
            assert "" == err

    @pytest.mark.parametrize("echo", [False, True], ids=["no_echo", "echo"])
    def test_run_timeout_try_terminate_then_kill(self, echo, capfd):
        command = "trap '' TERM; sleep 3"
        timeout = 1
        with pytest.raises(TimeoutExpired):
            run(command, shell=True, timeout=timeout, echo=echo, sigterm_timeout=1)

        if echo:
            out, err = capfd.readouterr()
            assert get_prompt_and_command(command, False, timeout) + "\n" == out
            assert "" == err

    @pytest.mark.parametrize("echo", [False, True], ids=["no_echo", "echo"])
    def test_run_timeout_and_kill(self, echo, capfd):
        command = "trap '' TERM; sleep 3"
        timeout = 1
        with pytest.raises(TimeoutExpired):
            run(command, shell=True, timeout=timeout, echo=echo, sigterm_timeout=0)

        if echo:
            out, err = capfd.readouterr()
            assert get_prompt_and_command(command, False, timeout) + "\n" == out
            assert "" == err

    @pytest.mark.parametrize("echo", [False, True], ids=["no_echo", "echo"])
    def test_run_timeout_and_terminate_sudo(self, echo, capfd):
        command = ["sudo", "sudo", "sh", "-c",
                   "sudo sleep 30 | $(cat | (dd 2>/dev/null) | cat)"]
        timeout = 3
        with pytest.raises(TimeoutExpired):
            run(command, timeout=timeout, echo=echo)

        if echo:
            out, err = capfd.readouterr()
            assert get_prompt_and_command(command, False, timeout) + "\n" == out
            assert "" == err

    @pytest.mark.parametrize("echo", [False, True], ids=["no_echo", "echo"])
    def test_run_timeout_try_terminate_then_kill_sudo(self, echo, capfd):
        command = "trap '' TERM; sudo sleep 30"
        timeout = 3
        with pytest.raises(TimeoutExpired):
            run(command, shell=True, timeout=timeout, echo=echo, sigterm_timeout=1)

        if echo:
            out, err = capfd.readouterr()
            assert get_prompt_and_command(command, False, timeout) + "\n" == out
            assert "" == err

    @pytest.mark.parametrize("echo", [False, True], ids=["no_echo", "echo"])
    def test_run_timeout_and_kill_sudo(self, echo, capfd):
        command = "trap '' TERM; sudo sleep 30"
        timeout = 3
        with pytest.raises(TimeoutExpired):
            run(command, shell=True, timeout=timeout, echo=echo, sigterm_timeout=0)

        if echo:
            out, err = capfd.readouterr()
            assert get_prompt_and_command(command, False, timeout) + "\n" == out
            assert "" == err

    @pytest.mark.parametrize("text", [False, True], ids=["binary", "text"])
    @pytest.mark.parametrize("echo", [False, True], ids=["no_echo", "echo"])
    def test_run_input(self, echo, text, capfd):
        if text:
            encoder = decoder = lambda x: x
        else:
            encoder = str.encode
            decoder = bytes.decode
        command = "cat"
        input = "Some input"
        output = run(
            command,
            echo=echo,
            input=encoder(input),
            text=text,
            capture_output=True
        )

        assert output.stdout == encoder(input)
        assert output.stderr == encoder("")
        assert output.returncode == 0
        assert output.time > 0

        out, err = capfd.readouterr()
        if echo:
            assert get_prompt_and_command(command) + f"\n{decoder(output.stdout)}" == out
            assert output.stderr == encoder(err)
        else:
            assert "" == out == err

    @pytest.mark.parametrize("manual_echo", [False, True], ids=["no_manual_echo", "manual_echo"])
    @pytest.mark.parametrize("extra_tees", [0, 1, 3], ids=["no_extra", "one_extra", "few_extra"])
    @pytest.mark.parametrize("stderr", [None, STDOUT], ids=["no_redirect", "redirect"])
    @pytest.mark.parametrize("stdout", [None, PIPE, FILE], ids=["no_capture", "pipe", "file"])
    @pytest.mark.parametrize("echo", [False, True], ids=["no_echo", "echo"])
    def test_run_tee(self, echo, stdout, stderr, extra_tees, manual_echo, capfd):
        redirect = stderr is STDOUT
        stdout_command = "echo This goes to stdout"
        stderr_command = "echo This goes to stderr >&2"

        count = extra_tees - 1 if extra_tees > 1 else extra_tees
        prompt_tees = {SpooledTemporaryFile(mode="w+t") for _ in range(count)}
        stdout_tees = {SpooledTemporaryFile(mode="w+t") for _ in range(count)}
        stderr_tees = {SpooledTemporaryFile(mode="w+t") for _ in range(count)}
        all_tees = {SpooledTemporaryFile(mode="w+t") for _ in range(extra_tees - count)}

        # Ensure we are sending the intended number of files
        assert extra_tees == len(prompt_tees) + len(all_tees)
        assert extra_tees == len(stdout_tees) + len(all_tees)
        assert extra_tees == len(stderr_tees) + len(all_tees)

        if manual_echo:
            prompt_tees.add(sys.stdout)
            stdout_tees.add(sys.stdout)
            stderr_tees.add(sys.stderr)

        kwargs = {
            "stdout": stdout,
            "stderr": stderr,
            "shell": True,
            "text": True,
            "echo": echo,
            "stdout_tees": stdout_tees | all_tees,
            "stderr_tees": stderr_tees | all_tees,
            "prompt_tees": prompt_tees | all_tees
        }

        expected_stdout_1 = expected_stderr_1 = expected_stdout_2 = expected_stderr_2 = None
        if stdout is not None:
            expected_stdout_1 = "This goes to stdout\n"
            expected_stdout_2 = "This goes to stderr\n" if redirect else ""

        count = 30
        for _ in range(count):
            output = run(stdout_command, **kwargs)
            assert expected_stdout_1 == output.stdout
            assert expected_stderr_1 == output.stderr

            output = run(stderr_command, **kwargs)
            assert expected_stdout_2 == output.stdout
            assert expected_stderr_2 == output.stderr

        expected_stdout_command_prompt = get_prompt_and_command(stdout_command, redirect) + "\n"
        expected_stderr_command_prompt = get_prompt_and_command(stderr_command, redirect) + "\n"
        expected_prompt = expected_stdout_command_prompt + expected_stderr_command_prompt
        expected_output = (
            expected_stdout_command_prompt + "This goes to stdout\n" +
            expected_stderr_command_prompt + "This goes to stderr\n"
        )

        if redirect:
            expected_stdout = "This goes to stdout\nThis goes to stderr\n"
            expected_stderr = ""
            if echo or manual_echo:
                expected_captured_stdout = expected_output
            else:
                expected_captured_stdout = expected_stdout if stdout is None else ""
        else:
            expected_stdout = "This goes to stdout\n"
            expected_stderr = "This goes to stderr\n"
            if echo or manual_echo:
                expected_captured_stdout = (
                    expected_stdout_command_prompt + "This goes to stdout\n" +
                    expected_stderr_command_prompt
                )
            else:
                expected_captured_stdout = expected_stdout if stdout is None else ""

        expected_content = expected_stdout * count
        for file in stdout_tees - {sys.stdout}:
            assert expected_content == read_file(file)

        expected_content = expected_stderr * count
        for file in stderr_tees - {sys.stderr}:
            assert expected_content == read_file(file)

        expected_content = expected_prompt * count
        for file in prompt_tees - {sys.stdout}:
            assert expected_content == read_file(file)

        expected_content = expected_output * count
        for file in all_tees:
            assert expected_content == read_file(file)

        out, err = capfd.readouterr()
        assert expected_captured_stdout * count == out
        assert expected_stderr * count == err

    @pytest.mark.parametrize("echo", [False, True], ids=["no_echo", "echo"])
    def test_run_alias(self, echo, capfd):
        command = ["echo", "-ne", "Hello!\\tBye."]
        alias = ["echo", "Hello!\\tBye."]
        output = run(command, echo=echo, alias=alias, **KWARGS)

        assert output.stdout == "Hello!\tBye."
        assert output.stderr == ""
        assert output.returncode == 0
        assert output.time > 0

        out, err = capfd.readouterr()
        if echo:
            assert get_prompt_and_command(alias) + f"\n{output.stdout}" == out
            assert output.stderr == err
        else:
            assert "" == out == err

    @patch("iripau.subprocess.run")
    def test_call(self, mock_run):
        returncode = call("arg1", "arg2", kwarg1="kwarg1")

        mock_run.assert_called_once_with("arg1", "arg2", kwarg1="kwarg1")
        assert mock_run.return_value.returncode == returncode

    @patch("iripau.subprocess.run")
    def test_check_call(self, mock_run):
        check_call("arg1", "arg2", kwarg1="kwarg1")

        mock_run.assert_called_once_with(
            "arg1",
            "arg2",
            kwarg1="kwarg1",
            check=True
        )

    @patch("iripau.subprocess.run")
    def test_check_output(self, mock_run):
        stdout = check_output("arg1", "arg2", kwarg1="kwarg1")

        mock_run.assert_called_once_with(
            "arg1",
            "arg2",
            kwarg1="kwarg1",
            check=True,
            stdout=PIPE
        )
        assert mock_run.return_value.stdout == stdout

    @patch("iripau.subprocess.run")
    def test_getoutput(self, mock_run):
        stdout = getoutput("arg1", "arg2", kwarg1="kwarg1")

        mock_run.assert_called_once_with(
            "arg1",
            "arg2",
            kwarg1="kwarg1",
            stdout=PIPE,
            stderr=STDOUT,
            shell=True,
            text=True
        )
        assert mock_run.return_value.stdout == stdout

    @patch("iripau.subprocess.run")
    def test_getstatusoutput(self, mock_run):
        returncode, stdout = getstatusoutput("arg1", "arg2", kwarg1="kwarg1")

        mock_run.assert_called_once_with(
            "arg1",
            "arg2",
            kwarg1="kwarg1",
            stdout=PIPE,
            stderr=STDOUT,
            shell=True,
            text=True
        )
        assert mock_run.return_value.returncode == returncode
        assert mock_run.return_value.stdout == stdout



"""
Tests to validate iripau.shutil module
"""

import os
import pytest
import shutil
import tempfile

from iripau.shutil import FileLock
from iripau.shutil import create_file
from iripau.shutil import read_file
from iripau.shutil import remove_file
from iripau.shutil import remove_tree
from iripau.shutil import file_created
from iripau.shutil import rotate


class TestFileLock:

    @classmethod
    def setup_class(cls):
        cls.file_name = "a_file.lock"

    def teardown_method(self):
        try:
            os.remove(self.file_name)
        except FileNotFoundError:
            pass

    def test_interfering(self):
        lock_1 = FileLock(self.file_name)
        lock_2 = FileLock(self.file_name)

        assert not lock_1.acquired
        assert not lock_2.acquired

        lock_1.acquire()
        assert lock_1.acquired
        assert not lock_2.acquired

        with pytest.raises(TimeoutError):
            lock_2.acquire(timeout=2)

        assert lock_1.acquired
        assert not lock_2.acquired

        lock_1.release()
        assert not lock_1.acquired
        assert not lock_2.acquired

        lock_2.acquire()
        assert not lock_1.acquired
        assert lock_2.acquired

        lock_2.release()
        assert not lock_1.acquired
        assert not lock_2.acquired

    def test_context(self):
        with FileLock(self.file_name):
            assert os.path.exists(self.file_name)
        assert not os.path.exists(self.file_name)


class TestShutil:

    def setup_method(self):
        self.workspace = tempfile.mkdtemp()

    def teardown_method(self):
        shutil.rmtree(self.workspace)

    def path(self, *paths):
        return os.path.join(self.workspace, *paths)

    def create_tree(self, *paths_contents):
        for path, content in paths_contents:
            path = self.path(path)
            dir_name, file_name = os.path.split(path)
            os.makedirs(dir_name, exist_ok=True)
            if file_name:
                create_file(path, content)

    def assert_tree(self, root, *paths_contents):
        for path, content in paths_contents:
            path = self.path(path)
            dir_name, file_name = os.path.split(path)
            if file_name:
                assert content == read_file(path)
            else:
                assert os.path.isdir(dir_name)

    @pytest.mark.parametrize("content", ["text", b"bytes"])
    def test_file_manipulation(self, content):
        file_name = self.path("a_file.tmp")
        create_file(file_name, content=content)
        assert content == read_file(file_name, bytes == type(content))
        remove_file(file_name)
        with pytest.raises(FileNotFoundError):
            os.remove(file_name)

    def test_remove_file_not_found(self):
        file_name = self.path("a_file.tmp")
        remove_file(file_name)

    def test_remove_tree_dir(self):
        self.create_tree(
            ("dir_1/dir_2/a_file.tmp", "this is a test!")
        )
        root = self.path("dir_1")
        remove_tree(root)
        assert not os.path.exists(root)

    def test_remove_tree_file(self):
        file_name = self.path("a_file.tmp")
        create_file(file_name)
        remove_tree(file_name)
        assert not os.path.exists(file_name)

    def test_remove_tree_dir_symlink(self):
        self.create_tree(
            ("dir_1/dir_2/a_file.tmp", "this is a test!")
        )
        root = self.path("dir_1")
        link = self.path("dir_link")
        os.symlink(root, link, target_is_directory=True)
        remove_tree(link)
        assert not os.path.exists(link)
        assert os.path.exists(root)

    def test_remove_tree_file_symlink(self):
        file_name = self.path("a_file.tmp")
        create_file(file_name)
        link = self.path("file_link")
        os.symlink(file_name, link)
        remove_tree(link)
        assert not os.path.exists(link)
        assert os.path.exists(file_name)

    def test_remove_tree_not_found(self):
        root = self.path("a_directory")
        remove_tree(root)

    def test_file_created(self):
        file_name = self.path("a_file.tmp")
        with file_created(file_name, content=file_name):
            assert file_name == read_file(file_name)
        with pytest.raises(FileNotFoundError):
            os.remove(file_name)

    def test_rotate(self, tmpdir):
        self.create_tree(
            ("dir_1/a_file.tmp", ""),
            ("dir_2/other_file_3.tmp", ""),
            ("dir_2.1/other_file_2.tmp", ""),
            ("dir_2.2/other_file_1.tmp", ""),
            ("dir_3/another_file.tmp", "EEE"),
            ("dir_3/another_file.1.tmp", "DDD"),
            ("dir_3/another_file.2.tmp", "CCC"),
            ("dir_3/another_file.4.tmp", "AAA")
        )

        rotate(self.path("dir_1"))
        rotate(self.path("dir_2"))
        rotate(self.path("dir_3/another_file.tmp"))

        self.assert_tree(
            ("dir_1.1/a_file.tmp", ""),
            ("dir_2.1/other_file_3.tmp", ""),
            ("dir_2.2/other_file_2.tmp", ""),
            ("dir_2.3/other_file_1.tmp", ""),
            ("dir_3/another_file.1.tmp", "EEE"),
            ("dir_3/another_file.2.tmp", "DDD"),
            ("dir_3/another_file.3.tmp", "CCC"),
            ("dir_3/another_file.4.tmp", "AAA")
        )
        assert not os.path.exists(self.path("dir_1"))
        assert not os.path.exists(self.path("dir_2"))
        assert not os.path.exists(self.path("dir_3/another_file.tmp"))



"""
Handle random data
"""

import math
import random
import string


def one(items):
    """ Return a new list containing only one element from items """
    return [random.choice(items)]


def some(items, percentage=50, at_least=2, at_most=None):
    """ Return a new list containing some elements from items.
        Ordinality is preserved as much as possible.
    """
    count = math.ceil(percentage * len(items) / 100)
    if at_least and at_least > count:
        count = at_least
    if at_most and at_most < count:
        count = at_most
    return sorted(random.sample(items, count), key=items.index)


def shuffled(items):
    """ Return a new list containing all of the elements from items but in a
        different order.
    """
    return random.sample(items, len(items))


def random_string(length, chars=string.ascii_letters):
    """ Return a string of the desired length containing randomly chosen
        characters from chars.
    """
    return "".join(random.choice(chars) for _ in range(length))



"""
Tests to validate iripau.threading module
"""

import pytest
import random
import threading

from time import sleep
from itertools import groupby
from multiprocessing.dummy import Pool

from iripau.threading import AsyncResult
from iripau.threading import MultiDequeuer
from iripau.threading import synchronized
from iripau.threading import cached


class TestAsyncResult:

    def test_success(self):
        def some_function(arg1, arg2=None):
            sleep(3)
            return arg1 + arg2

        result = AsyncResult(some_function, 1, 2)
        result.wait(1)
        assert not result.ready()
        with pytest.raises(ValueError):
            result.successful()
        with pytest.raises(TimeoutError):
            result.get(1)
        assert 3 == result.get()
        assert result.ready()
        assert result.successful()

    def test_success_context(self):
        def some_function(arg1, arg2=None):
            sleep(4)
            return arg1 + arg2

        with AsyncResult(some_function, 1, 2) as result:
            result.wait(1)
            assert not result.ready()
            with pytest.raises(ValueError):
                result.successful()
            with pytest.raises(TimeoutError):
                result.get(1)
        assert result.ready()
        assert 3 == result.get()
        assert result.successful()

    def test_exception(self):
        def some_function(arg1, arg2=None):
            sleep(3)
            assert False

        result = AsyncResult(some_function, 1)
        result.wait(1)
        assert not result.ready()
        with pytest.raises(ValueError):
            result.successful()
        with pytest.raises(TimeoutError):
            result.get(1)
        with pytest.raises(AssertionError):
            result.get()
        assert result.ready()
        assert not result.successful()


class TestMultiDequeuer:

    def test_consume_immediately(self):
        num_products = 10
        num_threads = num_products
        data = []

        def consumer(products):
            data.append(products)

        md = MultiDequeuer(consumer, collection_type=set)

        with Pool(num_threads) as pool:
            results = []
            for i in range(num_products):
                sleep(0.1)
                results.append(pool.apply_async(md.put, (i,)))
            [result.get() for result in results]

        assert [{0}, {1}, {2}, {3}, {4}, {5}, {6}, {7}, {8}, {9}] == data

    def test_consume_after_time(self):
        num_products = 10
        num_threads = num_products
        data = []

        def consumer(products):
            data.append(products)

        md = MultiDequeuer(consumer, 0.7, collection_type=set)

        with Pool(num_threads) as pool:
            results = []
            for i in range(num_products):
                sleep(0.1)
                results.append(pool.apply_async(md.put, (i,)))
            [result.get() for result in results]

        assert [{0, 1, 2, 3, 4, 5, 6}, {7, 8, 9}] == data

    def test_consume_when_count_reached(self):
        num_products = 10
        num_threads = num_products
        data = []

        def consumer(products):
            data.append(products)

        md = MultiDequeuer(consumer, None, 5, set)

        with Pool(num_threads) as pool:
            results = []
            for i in range(num_products):
                sleep(0.1)
                results.append(pool.apply_async(md.put, (i,)))
            [result.get() for result in results]

        assert [{0, 1, 2, 3, 4}, {5, 6, 7, 8, 9}] == data

    def test_consume_mixed(self):
        num_products = 10
        num_threads = num_products
        data = []

        def consumer(products):
            data.append(products)

        md = MultiDequeuer(consumer, 0.5, 4, set)

        with Pool(num_threads) as pool:
            results = []
            for i in range(num_products):
                sleep(0.1)
                results.append(pool.apply_async(md.put, (i,)))
            [result.get() for result in results]

        assert [{0, 1, 2, 3}, {4, 5, 6, 7}, {8, 9}] == data


class TestThreading:

    @pytest.mark.parametrize("cases", ["no_parenthesis", "empty_parenthesis", "parenthesis"])
    def test_synchronized(self, cases):
        num_inserts = 500
        num_threads = 500
        data = []

        match cases:
            case "no_parenthesis":
                decorator = synchronized
            case "empty_parenthesis":
                decorator = synchronized()
            case "parenthesis":
                decorator = synchronized(threading.BoundedSemaphore())

        @decorator
        def f(arg):
            for i in range(num_inserts):
                data.append(arg)

        with Pool(num_threads) as pool:
            results = [pool.apply_async(f, (i,)) for i in range(num_threads)]
            [result.get() for result in results]

        assert len(list(groupby(data))) == num_threads
        assert len(data) == num_threads * num_inserts

    @pytest.mark.parametrize("shifted", [True, False], ids=["no_parenthesis", "parenthesis"])
    def test_cached(self, shifted):
        num_inserts = 50
        num_repeats = 40
        num_threads = num_inserts * num_repeats
        data = []

        decorator = cached if shifted else cached()

        @decorator
        def f(arg):
            sleep(random.random())
            data.append(arg)
            return len(data)

        expected_data = list(range(num_inserts))
        expected_values = [
            i + 1
            for i in range(num_inserts)
            for j in range(num_repeats)
        ]

        with Pool(num_threads) as pool:
            results = [
                pool.apply_async(f, (i,))
                for i in expected_data
                for j in range(num_repeats)
            ]
            values = [result.get() for result in results]

        assert expected_values != values
        assert expected_data != data

        assert expected_values == sorted(values)
        assert expected_data == sorted(data)

    def test_cached_synchronized(self):
        num_inserts = 25
        num_repeats = 40
        num_threads = num_inserts * num_repeats
        data = []

        @cached(synchronized=True)
        def f(arg):
            sleep(random.random())
            data.append(arg)
            return len(data)

        expected_data = list(range(num_inserts))
        expected_values = [
            i + 1
            for i in range(num_inserts)
            for j in range(num_repeats)
        ]

        with Pool(num_threads) as pool:
            results = [
                pool.apply_async(f, (i,))
                for i in expected_data
                for j in range(num_repeats)
            ]
            values = [result.get() for result in results]

        assert expected_values == values
        assert expected_data == data



"""
Tests to validate iripau.random module
"""

import pytest
import random

from iripau.random import one
from iripau.random import some
from iripau.random import shuffled
from iripau.random import random_string


class TestRandom(object):

    def test_one(self):
        items = range(10)
        sample = one(items)
        assert 1 == len(sample)
        assert all(item in items for item in sample)

    @pytest.mark.parametrize("percentage", range(0, 101, 10))
    def test_some(self, percentage):
        items = range(100)
        sample = some(items, percentage=percentage, at_least=30, at_most=70)
        length = 30 if percentage < 30 else 70 if percentage > 70 else percentage
        assert length == len(sample)
        assert all(item in items for item in sample)
        assert sorted(sample) == sample

    def test_shuffled(self):
        items = [1, 2, 3, 4, 5, 6, 7, 8, 9, 0]
        shuffled_items = shuffled(items)
        assert items != shuffled_items
        assert len(items) == len(shuffled_items)
        for item in items:
            assert item in shuffled_items

    def test_random_string(self):
        length = random.randint(0, 20)
        assert length == len(random_string(length))



"""
Threading utilities
"""

import sys
import threading

from functools import wraps, _make_key as make_key
from collections import deque


class AsyncResult(threading.Thread):
    """ Implementation of the multiprocessing.pool.AsyncResult class but
        spawning a new thread to execute the call instead of using a Process or
        Thread Pool.

        https://docs.python.org/3/library/multiprocessing.html#multiprocessing.pool.AsyncResult
    """

    def __init__(self, function, *args, **kwargs):
        """ Call the function passing *args and **kwargs in a new thread """
        super().__init__(
            target=function,
            args=args,
            kwargs=kwargs
        )

        self.return_value = None
        self.exc_info = None

        self.start()

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.wait()

    def run(self):
        try:
            self.return_value = self._target(*self._args, **self._kwargs)
        except:  # noqa: E722
            self.exc_info = sys.exc_info()

    def get(self, timeout=None):
        self.join(timeout)
        if self.is_alive():
            raise TimeoutError("Timeout reached while joining the thread")
        if self.exc_info:
            exc_type, exc_value, traceback = self.exc_info
            raise exc_value.with_traceback(traceback)
        return self.return_value

    def wait(self, timeout=None):
        self.join(timeout)

    def ready(self):
        return not self.is_alive()

    def successful(self):
        if self.is_alive():
            raise ValueError("Thread has not completed")
        try:
            self.get()
        except:  # noqa: E722
            return False
        else:
            return True


class MultiDequeuer:
    """ Implementation of the Producer-Consumer Design Pattern in which one of
        the Producer workers will act as Consumer, processing all of the
        elements that have been produced so far. Also, the workers will block
        until the elements they produced are consumed.

        This is helpful to keep the logic of processing one single element per
        worker but consuming several elements more efficiently at once, without
        the need of having a separated thread or threads for consumption.

        If there is not any advantage on consuming multiple elements at once
        compared to consuming them one by one, it is preferred to consume the
        elements in the workers, keeping the logic of blocking the workers.

        If blocking the workers is nor desired, the regular Producer-Consumer
        Pattern will suffice.

        This class relies on the Python GIL.
    """

    def __init__(self, consumer, time_to_consume=None, count_hint=0, collection_type=list):
        """ The elements queue will be consumed by the consumer callable, which
            will be called by one of the workers. When the worker has started
            consuming the queue, a new queue will be created to handle future
            elements.

            If time_to_consume is greater than 0, the queue will be consumed
            when this time (in seconds) has passed after the first element was
            added to the queue.

            If count_hint is greater than 0, the queue will be consumed when at
            least this amount of elements have been added to the queue. Ignoring
            time_to_consume.

            The elements queue will be cast into collection_type before it is
            passed to the consumer callable.
        """
        self.consume = consumer
        self.count = count_hint
        self.time = time_to_consume
        self.type = collection_type

        self.lock = threading.Lock()
        self.data = self._new_data()

    def _new_data(self):
        return (
            deque(),
            threading.Barrier(self.count, timeout=self.time),
            threading.Lock()
        )

    def put(self, product):
        """ Add an element (product) to the queue. This method is meant to be
            called by the workers. It blocks until the element is consumed.
        """
        data = self.data

        queue, barrier, lock = data

        queue.append(product)
        if self.count or self.time:
            try:
                barrier.wait()
            except threading.BrokenBarrierError:
                pass

        with self.lock:
            if self.data is data:
                self.data = self._new_data()

        with lock:
            if queue:
                products = self.type(queue)
                self.consume(products)
                queue.clear()


class FunctionCacher:
    """ Cache the return value of a callable depending on their call arguments.
        The first call will generate the cache and subsequent calls will return
        the cached value without executing the original callable.

        The instance of this class is meant to be called as if it were the
        original callable.

        The main difference from the functools.cache decorator is that the cache
        for a particular set of arguments will be generated by only one
        execution of the original callable; if more than one thread calls this
        object before the cache is generated, only one of them will actually
        execute the original callable and the rest will wait for it to finish.

        https://docs.python.org/3/library/functools.html#functools.cache
    """

    def __init__(self, function, synchronized=False, enabled=True):
        """ The callable to be cached is function.

            If enabled is False, caching will start disabled. It can be enabled
            later on.

            If synchronized is True, the original callable will be protected
            with a lock to prevent it is executed in parallel.
        """
        self.function = function
        if enabled:
            self.enable_cache(synchronized)
        else:
            self.disable_cache()
        self.lock = threading.Lock()

    def __call__(self, *args, **kwargs):
        return self.callable(*args, **kwargs)

    def enable_cache(self, synchronized=False):
        """ Enable caching if not enabled already.
            The synchronized argument is used as in the constructor.
            This method is not re-entrant nor thread-safe.
        """
        if not hasattr(self, "cache"):
            self.cache = {}
        if synchronized:
            if hasattr(self, "locks"):
                del self.locks
            self.callable = self._synced
        else:
            if not hasattr(self, "locks"):
                self.locks = {}
            self.callable = self._cached

    def disable_cache(self):
        """ Disable and delete caching if not disabled already.
            This method is not re-entrant nor thread-safe.
        """
        if hasattr(self, "cache"):
            del self.cache
        if hasattr(self, "locks"):
            del self.locks
        self.callable = self.function

    def clear_cache(self):
        """ Delete cache if caching is not disabled.
            This method is not re-entrant nor thread-safe.
        """
        if hasattr(self, "cache"):
            self.cache = {}
        if hasattr(self, "locks"):
            self.locks = {}

    def _cached(self, *args, **kwargs):
        key = make_key(args, kwargs, False)
        if key not in self.cache:
            with self.lock:
                if key not in self.locks:
                    self.locks[key] = threading.Lock()
            with self.locks[key]:
                if key not in self.cache:
                    self.cache[key] = self.function(*args, **kwargs)
        return self.cache[key]

    def _synced(self, *args, **kwargs):
        key = make_key(args, kwargs, False)
        if key not in self.cache:
            with self.lock:
                if key not in self.cache:
                    self.cache[key] = self.function(*args, **kwargs)
        return self.cache[key]


def synchronized(lock=None):
    """ Decorator to wrap a function with a lock/semaphore """
    if callable(lock):
        function, lock = lock, threading.Lock()
        shifted = True
    else:
        if lock is None:
            lock = threading.Lock()
        shifted = False

    def decorator(function):

        @wraps(function)
        def wrapper(*args, **kwargs):
            with lock:
                return function(*args, **kwargs)

        return wrapper

    return decorator(function) if shifted else decorator


def cached(synchronized=False, enabled=True):
    """ Decorator using the FunctionCacher class. """
    if callable(synchronized):
        function, synchronized = synchronized, False
        shifted = True
    else:
        shifted = False

    def decorator(function):
        return wraps(function)(FunctionCacher(function, synchronized, enabled))

    return decorator(function) if shifted else decorator



"""
Tests to validate iripau.command module
"""

import os
import pytest

from mock import patch
from shlex import quote
from socket import gethostname
from getpass import getuser

from iripau.command import _solve_ssh_users
from iripau.command import user_cmd
from iripau.command import local_run
from iripau.command import local_run_interactive
from iripau.command import ssh_run
from iripau.command import ssh_run_interactive
from iripau.command import host_run
from iripau.command import host_run_interactive

from iripau.subprocess import DEVNULL

ROWS = os.popen("tput lines").read().rstrip()
COLS = os.popen("tput cols").read().rstrip()
STTY_CMD = "stty rows {0} cols {1} && ".format(ROWS, COLS)
HOSTNAME = gethostname()
USER = getuser()
KWARGS = {"kwarg1": "arg1", "kwarg2": 2, "kwarg3": "Sup!", "kwarg4": 4}


def run_kwargs(
    args, *, executable=None, stdin=None, stdout=None, stderr=None, shell=False,
    cwd=None, env=None, user=None, encoding=None, errors=None, text=True,
    stdout_tees=[], add_global_stdout_tees=True,
    stderr_tees=[], add_global_stderr_tees=True,
    prompt_tees=[], add_global_prompt_tees=True,
    echo=None, alias=None, input=None, capture_output=False,
    timeout=120, check=False, sigterm_timeout=10
):
    return locals()


class TestCommand:

    @patch("iripau.command.USER", new="current-user")
    def test_solve_ssh_users(self):
        assert ("current-user", "host1") == \
            _solve_ssh_users("host1", None)

        assert ("current-user", "host1") == \
            _solve_ssh_users("host1", "current-user")

        assert ("other-user", "host1") == \
            _solve_ssh_users("host1", "other-user")

        assert ("current-user", "host1") == \
            _solve_ssh_users("current-user@host1", None)

        assert ("current-user", "host1") == \
            _solve_ssh_users("current-user@host1", "current-user")

        assert ("current-user", "current-user@host1") == \
            _solve_ssh_users("current-user@host1", "other-user")

        assert ("other-user", "other-user@host1") == \
            _solve_ssh_users("other-user@host1", None)

        assert ("other-user", "other-user@host1") == \
            _solve_ssh_users("other-user@host1", "current-user")

        assert ("other-user", "host1") == \
            _solve_ssh_users("other-user@host1", "other-user")

    @pytest.mark.parametrize("env", [None, {"FOO": "foo", "BAR": "bar"}], ids=["no_env", "env"])
    @pytest.mark.parametrize("alias_type", [None, list, str])
    @pytest.mark.parametrize("cmd_type", [list, str])
    @pytest.mark.parametrize("user", ["no_user", "current-user", "other-user", "root"])
    @patch("iripau.command.USER", new="current-user")
    def test_user_cmd(self, user, cmd_type, alias_type, env):
        if user == "no_user":
            user = None

        env_tokens = list(map("{0[0]}={0[1]}".format, (env or {}).items()))
        alias = {
            str: "echo bye",
            list: ["echo", "bye"],
            None: None
        }[alias_type]
        if cmd_type is str:
            cmd = "whoami; sleep 1 && echo Bye!"
            internal_alias = {
                str: ["sh", "-c", alias],
                list: alias,
                None: ["sh", "-c", cmd]
            }[alias_type]
            expected_cmds = {
                None: (
                    cmd,
                    alias
                ),
                "current-user": (
                    cmd,
                    alias
                ),
                "other-user": (
                    ["sudo", "-Eu", user] + env_tokens + ["sh", "-c", cmd],
                    ["sudo", "-Eu", user] + internal_alias
                ),
                "root": (
                    ["sudo", "-E"] + env_tokens + ["sh", "-c", cmd],
                    ["sudo", "-E"] + internal_alias
                )
            }
        else:
            cmd = ["echo", "Hello!", "Bye!"]
            internal_alias = {
                str: ["sh", "-c", alias],
                list: alias,
                None: cmd
            }[alias_type]
            expected_cmds = {
                None: (
                    cmd,
                    alias
                ),
                "current-user": (
                    cmd,
                    alias
                ),
                "other-user": (
                    ["sudo", "-Eu", user] + env_tokens + cmd,
                    ["sudo", "-Eu", user] + internal_alias
                ),
                "root": (
                    ["sudo", "-E"] + env_tokens + cmd,
                    ["sudo", "-E"] + internal_alias
                )
            }
        assert expected_cmds[user] == user_cmd(user, cmd, alias, env)

    @pytest.mark.parametrize("env", [None, {"FOO": "foo", "BAR": "bar"}], ids=["no_env", "env"])
    @pytest.mark.parametrize("cmd_type", [list, str])
    @pytest.mark.parametrize("user", ["no_user", "root", "no-root"])
    @patch("iripau.command.USER", new="root")
    def test_user_cmd_being_root(self, user, cmd_type, env):
        if user == "no_user":
            user = None

        env_tokens = list(map("{0[0]}={0[1]}".format, (env or {}).items()))
        if cmd_type is str:
            cmd = "whoami; sleep 1 && echo Bye!"
            expected_cmds = {
                None: (cmd, None),
                "root": (cmd, None),
                "no-root": (
                    ["sudo", "-Eu", user] + env_tokens + ["sh", "-c", cmd],
                    ["sudo", "-Eu", user, "sh", "-c", cmd]
                )
            }
        else:
            cmd = ["echo", "Hello!", "Bye!"]
            expected_cmds = {
                None: (cmd, None),
                "root": (cmd, None),
                "no-root": (
                    ["sudo", "-Eu", user] + env_tokens + cmd,
                    ["sudo", "-Eu", user] + cmd
                )
            }
        assert expected_cmds[user] == user_cmd(user, cmd, None, env)

    @pytest.mark.parametrize("cmd_type", [list, str])
    @patch("iripau.subprocess.run")
    def test_local_run(self, mock_run, cmd_type):
        cmd = "echo Hello!" if cmd_type is str else ["echo", "Hello!"]

        output = local_run(cmd)

        mock_run.assert_called_once_with(
            **run_kwargs(cmd, shell=cmd_type is str, stdin=DEVNULL, capture_output=True)
        )
        assert mock_run.return_value == output

    @pytest.mark.parametrize("env", [None, {"FOO": "foo", "BAR": "bar"}], ids=["no_env", "env"])
    @pytest.mark.parametrize("cmd_type", [list, str])
    @patch("iripau.command.spawn")
    def test_local_run_interactive(self, mock_spawn, cmd_type, env):
        cmd = "bash -i" if cmd_type is str else ["bash", "-i"]

        result = local_run_interactive(cmd, env=env)

        if cmd_type is list:
            cmd = " ".join(quote(token) for token in cmd)
        cmd = STTY_CMD + cmd
        if env:
            env_tokens = list(map("{0[0]}={0[1]}".format, env.items()))
            cmd = "export " + " ".join(env_tokens) + " && " + cmd
        mock_spawn.assert_called_once_with(["sh", "-ic", cmd])
        assert mock_spawn.return_value == result

    @pytest.mark.parametrize("ssh_password", [None, "a_password"], ids=["no_ssh_pass", "ssh_pass"])
    @pytest.mark.parametrize("ssh_args", [[], ["-O", "exit"]], ids=["no_ssh_args", "ssh_args"])
    @pytest.mark.parametrize("env", [None, {"FOO": "foo", "BAR": "bar"}], ids=["no_env", "env"])
    @pytest.mark.parametrize("cwd", [None, "/some/path"], ids=["no_cwd", "cwd"])
    @pytest.mark.parametrize("cmd_type", [list, str])
    @patch("iripau.subprocess.run")
    def test_ssh_run(self, mock_run, cmd_type, cwd, env, ssh_args, ssh_password):
        host = "user@host"
        cmd = "echo Hello!" if cmd_type is str else ["echo", "Hello!"]

        output = ssh_run(
            host,
            cmd,
            cwd=cwd, env=env,
            ssh_args=ssh_args,
            ssh_password=ssh_password
        )

        if cmd_type is not str:
            cmd = " ".join(map(quote, cmd))
        alias = ["ssh", host, cmd]
        if cwd:
            cmd = "cd /some/path && " + cmd
        if env:
            cmd = "export FOO=foo BAR=bar && " + cmd
        ssh_cmd = ["ssh"] + ssh_args + [host, cmd]

        if ssh_password:
            ssh_cmd = ["sshpass", "-p", ssh_password] + ssh_cmd

        mock_run.assert_called_once_with(
            **run_kwargs(ssh_cmd, shell=False, alias=alias, stdin=DEVNULL, capture_output=True)
        )
        assert mock_run.return_value == output

    @pytest.mark.parametrize("ssh_password", [None, "a_password"], ids=["no_ssh_pass", "ssh_pass"])
    @pytest.mark.parametrize("ssh_args", [[], ["-O", "exit"]], ids=["no_ssh_args", "ssh_args"])
    @pytest.mark.parametrize("cmd_type", [list, str])
    @patch("iripau.command.spawn")
    def test_ssh_run_interactive(self, mock_spawn, cmd_type, ssh_args, ssh_password):
        host = "user@host"
        cmd = "bash -i" if cmd_type is str else ["bash", "-i"]

        result = ssh_run_interactive(
            host,
            cmd,
            ssh_args=ssh_args,
            ssh_password=ssh_password
        )

        if cmd_type is list:
            cmd = " ".join(quote(token) for token in cmd)
        ssh_cmd = ["ssh", "-tt"] + ssh_args + [host, STTY_CMD + cmd]

        if ssh_password:
            ssh_cmd = ["sshpass", "-p", ssh_password] + ssh_cmd

        mock_spawn.assert_called_once_with(ssh_cmd)
        assert mock_spawn.return_value == result

    @pytest.mark.parametrize("host", ["localhost", HOSTNAME])
    @patch("iripau.command.ssh_args")
    @patch("iripau.command.local_args", return_value=KWARGS)
    @patch("iripau.subprocess.run")
    def test_host_run_local(self, mock_run, mock_local_args, mock_ssh_args, host):
        command = "echo Hello!"

        output = host_run(command, host=host)

        mock_local_args.assert_called_once_with(command)
        mock_ssh_args.assert_not_called()
        mock_run.assert_called_once_with(**KWARGS)
        assert mock_run.return_value == output

    @pytest.mark.parametrize("host", ["localhost", HOSTNAME])
    @patch("iripau.command.local_run_interactive")
    def test_host_run_interactive_local(self, mock_run, host):
        command = "echo Hello!"

        result = host_run_interactive(command, host=host)

        mock_run.assert_called_once_with(command)
        assert mock_run.return_value == result

    @patch("iripau.command.ssh_args", return_value=KWARGS)
    @patch("iripau.command.local_args")
    @patch("iripau.subprocess.run")
    def test_host_run_remote(self, mock_run, mock_local_args, mock_ssh_args):
        host = "user@host"
        command = "echo Hello!"

        output = host_run(command, host=host)

        mock_local_args.assert_not_called()
        mock_ssh_args.assert_called_once_with(
            host,
            command,
            ssh_user=None,
            ssh_password=None,
            ssh_args=[],
            add_global_ssh_args=True
        )
        mock_run.assert_called_once_with(**KWARGS)
        assert mock_run.return_value == output

    @patch("iripau.command.ssh_run_interactive")
    def test_host_run_interactive_remote(self, mock_run):
        host = "user@host"
        command = "echo Hello!"

        result = host_run_interactive(command, host=host)

        mock_run.assert_called_once_with(host, command, ssh_user=None)
        assert mock_run.return_value == result



"""
Tests to validate iripau.logging module
"""

import os
import re
import pytest
import random
import logging
import tempfile
import threading

from time import sleep

from iripau.logging import LoggerFile
from iripau.logging import SimpleThreadNameFormatter
from iripau.logging import group_log_lines


class TestLogging:

    @staticmethod
    def write(stream, text):
        stream.write(text)
        stream.flush()

    @staticmethod
    def read(stream):
        sleep(0.1)
        stream.seek(0)
        return stream.read()

    def test_logger_file(self):
        output_file = tempfile.SpooledTemporaryFile(mode="w+t")

        # Setup log handler
        formatter = SimpleThreadNameFormatter("%(levelname)s: %(message)s")
        handler = logging.StreamHandler(output_file)
        handler.setLevel(logging.DEBUG)
        handler.setFormatter(formatter)

        # Setup logger
        logger = logging.getLogger("dummy_logger")
        logger.setLevel(logging.DEBUG)
        logger.addHandler(handler)

        logger_file = LoggerFile(logger, logging.INFO)

        # Write a part of a line
        self.write(logger_file, "Hello,")
        content = self.read(output_file)
        assert "" == content

        # Write the rest of the line
        self.write(logger_file, " bye\n")
        content = self.read(output_file)
        assert content.endswith(
            "INFO: Hello, bye\n"
        )

        # Write several lines
        self.write(
            logger_file,
            "SomeString\n"
            "AnotherString\n"
            "TestString\n"
            "DummyString\n"
        )

        content = self.read(output_file)
        assert content.endswith(
            "INFO: SomeString\n"
            "INFO: AnotherString\n"
            "INFO: TestString\n"
            "INFO: DummyString\n"
        )

        # Write a long line
        self.write(
            logger_file,
            "SomeString AnotherString TestString DummyString\n"
        )

        content = self.read(output_file)
        assert content.endswith(
            "INFO: SomeString AnotherString TestString DummyString\n"
        )

        # Write one line using file descriptor
        fd = logger_file.fileno()
        os.write(fd, b"Another dummy string\n")
        content = self.read(output_file)
        assert content.endswith(
            "INFO: Another dummy string\n"
        )

        # Write several lines using file descriptor
        self.write(
            logger_file,
            "Line1\n"
            "Line2\n"
            "Line3\n"
            "Line4\n"
        )

        logger_file.close()
        content = self.read(output_file)
        assert content.endswith(
            "INFO: Line1\n"
            "INFO: Line2\n"
            "INFO: Line3\n"
            "INFO: Line4\n"
        )

    @staticmethod
    def log_stuff(id, normal_logger, stdout_logger, stderr_logger):
        sleep(random.uniform(0.1, 0.3))
        normal_logger.info(f"{id} - First line")

        sleep(random.uniform(0.1, 0.3))
        stdout_logger.info(f"{id} - Simulated output")

        sleep(random.uniform(0.1, 0.3))
        stderr_logger.info(f"{id} - Simulated error")

        sleep(random.uniform(0.1, 0.3))
        normal_logger.info(f"{id} - Last line")

    @classmethod
    def log_stuff_in_threads(cls, ids, *loggers):
        threads = [
            threading.Thread(target=cls.log_stuff, args=(id, *loggers))
            for id in ids
        ]

        for thread in threads:
            thread.start()

        for thread in threads:
            thread.join()

    def test_group_log_lines(self):
        output_file = tempfile.SpooledTemporaryFile(mode="w+t")

        normal_log_format = "%(asctime).19s - %(threadName)12.12s: %(message)s"
        output_log_format = "%(asctime).19s - %(threadName)12.12s: [%(name)s] %(message)s"

        # Setup normal log handler
        formatter = SimpleThreadNameFormatter(normal_log_format)
        handler = logging.StreamHandler(output_file)
        handler.setLevel(logging.DEBUG)
        handler.setFormatter(formatter)

        # Setup normal logger
        logger = logging.getLogger(__name__)
        logger.setLevel(logging.DEBUG)
        logger.addHandler(handler)

        # Setup stdout and stderr log handler
        formatter = SimpleThreadNameFormatter(output_log_format)
        handler = logging.StreamHandler(output_file)
        handler.setLevel(logging.DEBUG)
        handler.setFormatter(formatter)

        # Setup stdout logger
        stdout_logger = logging.getLogger("stdout")
        stdout_logger.setLevel(logging.DEBUG)
        stdout_logger.addHandler(handler)

        # Setup stderr logger
        stderr_logger = logging.getLogger("stderr")
        stderr_logger.setLevel(logging.DEBUG)
        stderr_logger.addHandler(handler)

        self.log_stuff_in_threads((1, 2), logger, stdout_logger, stderr_logger)
        logger.info("Line from Main")
        self.log_stuff_in_threads((3, 4), logger, stdout_logger, stderr_logger)

        pattern = (
            ".{34}: (1|2) - First line\n"
            ".{34}: \\[stdout\\] \\1 - Simulated output\n"
            ".{34}: \\[stderr\\] \\1 - Simulated error\n"
            ".{34}: \\1 - Last line\n"
            ".{34}: (1|2) - First line\n"
            ".{34}: \\[stdout\\] \\2 - Simulated output\n"
            ".{34}: \\[stderr\\] \\2 - Simulated error\n"
            ".{34}: \\2 - Last line\n"
            ".{34}: Line from Main\n"
            ".{34}: (3|4) - First line\n"
            ".{34}: \\[stdout\\] \\3 - Simulated output\n"
            ".{34}: \\[stderr\\] \\3 - Simulated error\n"
            ".{34}: \\3 - Last line\n"
            ".{34}: (3|4) - First line\n"
            ".{34}: \\[stdout\\] \\4 - Simulated output\n"
            ".{34}: \\[stderr\\] \\4 - Simulated error\n"
            ".{34}: \\4 - Last line\n"
        )

        content = self.read(output_file)
        assert not re.fullmatch(pattern, content)  # Verify content is scrambled

        output_file.seek(0)
        grouped_lines = group_log_lines(output_file, r".* - \s*([^:\s]+).*: .*")
        assert re.fullmatch(pattern, "".join(grouped_lines))

    def test_group_log_lines_invalid(self):
        lines = ["Some log line\n"]
        with pytest.raises(ValueError):
            list(group_log_lines(lines, "unmatching regex"))



"""
Shell utilities
"""

import os
import errno
import select
import shutil
import contextlib

from iripau.functools import wait_for


class FileLock:
    poll_time = 0.05

    def __init__(self, file_name, timeout=None):
        self.file_name = file_name
        self.timeout = timeout
        self.acquired = False

    def __enter__(self):
        self.acquire(self.timeout)
        return self

    def __exit__(self, type, value, traceback):
        self.release()

    def __del__(self):
        self.release()

    def _lock_file_created(self):
        """ Lock file could be created by us """
        try:
            self.fd = os.open(
                self.file_name,
                os.O_CREAT | os.O_EXCL | os.O_RDWR
            )
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise
        else:
            return True
        return False

    def acquire(self, timeout=None):
        if not self.acquired:
            wait_for(
                self._lock_file_created,
                _timeout=timeout or self.timeout,
                _poll_time=self.poll_time
            )
            self.acquired = True

    def release(self):
        if self.acquired:
            os.close(self.fd)
            os.unlink(self.file_name)
            self.acquired = False


def create_file(file_name, content=""):
    mode = "wb" if bytes == type(content) else "w"
    with open(file_name, mode) as f:
        f.write(content)


def read_file(file_name, binary=False):
    mode = "rb" if binary else "r"
    with open(file_name, mode) as f:
        return f.read()


def remove_file(file_name):
    with contextlib.suppress(FileNotFoundError):
        os.remove(file_name)


def remove_tree(root):
    with contextlib.suppress(FileNotFoundError):
        if os.path.isdir(root) and not os.path.islink(root):
            shutil.rmtree(root)
        else:
            os.remove(root)


@contextlib.contextmanager
def file_created(file_name, content=""):
    create_file(file_name, content)
    try:
        yield file_name
    finally:
        os.remove(file_name)


def wait_for_file(file_obj, *args, **kwargs):
    poll_obj = select.poll()
    poll_obj.register(file_obj, select.POLLIN)
    wait_for(lambda: poll_obj.poll(0), *args, **kwargs)


def _rotate(root, ext, seq=0):
    old = seq and f"{root}.{seq}{ext}" or root + ext
    if os.path.exists(old):
        seq += 1
        _rotate(root, ext, seq)
        os.rename(old, f"{root}.{seq}{ext}")


def rotate(path, seq=0):
    _rotate(*os.path.splitext(path))



"""
Logging utilities
"""

import os
import re
import logging
import threading

from collections import OrderedDict


class LoggerFile:
    """ File that logs every line written to it """

    def __new__(cls, logger, level):
        r, w = os.pipe()
        read_file = os.fdopen(r, "r")
        write_file = os.fdopen(w, "w")

        thread = threading.Thread(
            target=cls.log,
            args=(read_file, logger, level),
            name=threading.current_thread().name
        )

        # Close write_file when the thread joins
        cls.patch(thread, "join", write_file.close)

        # Join the thread when the file is closed
        write_file.close = thread.join

        thread.start()
        return write_file

    @staticmethod
    def log(read_file, logger, level):
        with read_file:
            for line in read_file:
                logger.log(level, line.splitlines()[0])

    @staticmethod
    def patch(instance, function_name, callback):
        original_function = getattr(instance, function_name)

        def new_function():
            callback()
            original_function()
        setattr(instance, function_name, new_function)


class SimpleThreadNameFormatter(logging.Formatter):
    """ The same logging.Formatter but threadName is just the first token after
        splitting it: record.threadName.split(maxsplit=1)[0]
    """

    def format(self, record):
        if record.threadName:
            record.threadName = record.threadName.split(maxsplit=1)[0]
        return super().format(record)


def group_log_lines(lines, thread_id_regex, main_thread_id="MainThread"):
    """ For a log file containing entries from several threads, group the lines
        so that the lines coming from the same thread are contiguous, preserving
        order within the group. Logs coming from MainThread will not be grouped.
    """
    lines_map = OrderedDict()
    for i, line in enumerate(lines):
        match = re.match(thread_id_regex, line)
        if not match:
            raise ValueError(f"Invalid log line {i}: '{line}'")

        thread = match.groups()[0]
        if main_thread_id == thread:
            for thread_lines in lines_map.values():
                for thread_line in thread_lines:
                    yield thread_line
            yield line
            lines_map.clear()
        else:
            lines_map[thread] = lines_map.get(thread, []) + [line]

    for thread_lines in lines_map.values():
        for thread_line in thread_lines:
            yield thread_line
